/*
Fragerizer - Profile-guided load order optmization for Windows

Copyright (c) 2011, Matt Weeks 
All rights reserved. 

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met: 

 * Redistributions of source code must retain the above copyright notice, 
   this list of conditions and the following disclaimer. 
 * Redistributions in binary form must reproduce the above copyright 
   notice, this list of conditions and the following disclaimer in the 
   documentation and/or other materials provided with the distribution. 
 * Neither the name of  nor the names of its contributors may be used to 
   endorse or promote products derived from this software without specific 
   prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE 
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
POSSIBILITY OF SUCH DAMAGE. 
 */ 

//#define VERBOSE //uncomment this for more fun
#include <iostream>
#include <sstream>
#include <fstream>
#include <vector>
#include <string>
#include <set>
#include <map>
#include <math.h>
#include <windows.h>
using namespace std;

//Print windows error message
void printErr(){
	LPVOID lpMsgBuf;
	DWORD dw = GetLastError(); 
	FormatMessageW(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | 
		FORMAT_MESSAGE_IGNORE_INSERTS, NULL, dw, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		(LPWSTR) &lpMsgBuf, 0, NULL );
	wcerr << "System error " << dw << ": " << (PWCHAR)lpMsgBuf << endl;
	//LocalFree(lpMsgBuf);
}

//Exit printing message and windows error code info
void die(PCHAR message){
	cerr << message << endl;
	if(GetLastError() != ERROR_SUCCESS)
		printErr();
	exit(1);
}

//Find a contiguous set of clusters in a bitmap
LONGLONG findFreeClusterSet(PVOLUME_BITMAP_BUFFER bitmap, LONGLONG numClusters){
	DWORD bytesOfBitmapToReserve = numClusters / 8 + 1; //so we don't need to round up 
	//and it's easier working with bytes than bits
	LONGLONG sequential = 0, i;
	//Loop i until we hit the end of the disk, or have found enough contiguous space
	for(i = 0; i < bitmap->BitmapSize.QuadPart/8; i++){
		if(bitmap->Buffer[i] != 0){
			sequential = 0;
			continue;
		}
		sequential++;
		if(sequential > bytesOfBitmapToReserve)
			break; // The last numClusters clusters 
			//(beginning at bitmap buffer index i - bytesOfBitmapToReserve) are free, so break
	}
	if(i >= bitmap->BitmapSize.QuadPart/8)
		die("Cannot find enough free contiguous space!");

	//Oh good, we do have enough!
	LONGLONG currentDiskCluster = ((i - bytesOfBitmapToReserve)*8) + bitmap->StartingLcn.QuadPart;
	cout << "Found " << numClusters << " sequential free clusters at cluster " << currentDiskCluster << endl;
	return currentDiskCluster;
}

//Here we go!
int main (int argc, char ** argv){
	cout << "Fragerizer - profile-guided load-order disk optimizer" << endl
		<< "by scriptjunkie " << __DATE__ << endl << endl;
	if(argc < 3)
		die("USAGE: fragerizer ProcmonOutputFile.csv C:\\path\\to\\optimize");

	//Open up file
	ifstream inFile (argv[1]);
	string line;
	set<string> processedFiles;
	vector<string> files;
	map<string,DWORD> fileClusterCounts;
	DWORD clustersToReserve = 0;

	//Verify path to optimize
	char volumeName[MAX_PATH]; 
	BOOL result = GetVolumePathNameA(argv[2], volumeName, sizeof(volumeName));
	if (!result)
		die("Invalid path to optimize. :0");

	//Get basic disk info
	DWORD SectorsPerCluster, BytesPerSector, NumberOfFreeClusters, TotalNumberOfClusters;
	DWORD clusterSize;
	result = GetDiskFreeSpaceA(volumeName, 
				&SectorsPerCluster,
				&BytesPerSector,
				&NumberOfFreeClusters,
				&TotalNumberOfClusters);
	clusterSize = BytesPerSector * SectorsPerCluster;
	 if (!result)
		die("Can't get disk parameters. :0");

	 //Format name for volume handle, i.e. C:\ becomes \\.\C:
	 string deviceName;
	 deviceName.append("\\\\.\\");
	 deviceName.append(volumeName);
	 if(deviceName.at(deviceName.length()-1) == '\\')
		 deviceName.erase(deviceName.length() - 1);
	 //Open volume
	 HANDLE vol = CreateFileA(deviceName.c_str(), GENERIC_READ|GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE,0,OPEN_EXISTING,0,0);
	 if(vol == INVALID_HANDLE_VALUE)
		die("Can't open disk!");

	 //Get bitmap
	 DWORD bitmapSizeIn = TotalNumberOfClusters/8 + 17;
	 PVOLUME_BITMAP_BUFFER bitmap = (PVOLUME_BITMAP_BUFFER)VirtualAlloc(NULL, bitmapSizeIn, MEM_COMMIT, PAGE_READWRITE);
	 if(bitmap == NULL)
		die("Can't allocate memory for bitmap!");
	 STARTING_LCN_INPUT_BUFFER inputLCN;
	 inputLCN.StartingLcn.QuadPart = (long long)0;
	 DWORD bitmapSize;
	 result = DeviceIoControl(vol, FSCTL_GET_VOLUME_BITMAP, &inputLCN, sizeof(inputLCN), 
		 bitmap, bitmapSizeIn, &bitmapSize, NULL);
	 if(result != TRUE)
		die("Can't get volume bitmap!");

	//Iterate through procmon csv, getting ordered list of files and total number of clusters
	cout << "Building list..." << endl;
	while (getline (inFile, line)){
		istringstream linestream(line);
		vector<string> items;
		string item;
		//Turn CSV into vector
		while (line.length() > 0){
			if(line.at(0) == '"'){
				item = line.substr(1,line.find("\"",1)-1);
				if(line.length() <= item.length() + 3)
					line = "";
				else
					line = line.substr(item.length() + 3);
			}else{
				if(line.find(",") == -1){
					item = "";
					line = "";
				}else{
					item = line.substr(0,line.find(","));
					line = line.substr(line.find(",")+1);
				}
			}
			items.push_back(item);
		}
		//Find any successful events accessing/manipulating/creating files or directories
		if((items.at(3).find("File") == -1 && items.at(3).find("Directory") == -1) || items.at(5) != "SUCCESS" 
				|| items.at(4).find(argv[2]) == -1)
			continue; //not what we're looking for
		string fname = items.at(4);
		//Only new events
		if(processedFiles.find(fname) != processedFiles.end())
			continue;
		processedFiles.insert(fname); //make sure we don't move this again
		//Open file or directory and get size to figure out how many clusters will be needed
		LARGE_INTEGER fsize;
		HANDLE fileHandle = CreateFileA(fname.c_str(),FILE_READ_ATTRIBUTES,FILE_SHARE_READ | 
				FILE_SHARE_WRITE | FILE_SHARE_DELETE,0,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0);
		if(fileHandle == INVALID_HANDLE_VALUE || GetFileSizeEx(fileHandle,&fsize) == FALSE){
#ifdef VERBOSE
			//printErr();
			cerr << fname.c_str() << " ERROR Cannot get file size" << endl;//get file size
#endif
			CloseHandle(fileHandle);
			continue;
		}
		CloseHandle(fileHandle);
#ifdef VERBOSE
		cout << fname << endl; // Print out the file we found
#endif
		files.push_back(fname); // Add to our ordered list
		DWORD clusterCount = ceil(fsize.QuadPart / (clusterSize * 1.0)); // Get cluster count
		fileClusterCounts[fname] = clusterCount; // Save cluster count
		clustersToReserve += clusterCount; // Update total cluster count
	}

	//Find sequential free clusters
	cout << "Looking for " << clustersToReserve << " sequential free clusters for " 
			<< files.size() << " disk objects (files, directories)..." << endl;
	LONGLONG currentDiskCluster = findFreeClusterSet(bitmap,clustersToReserve);

	//Here we go!
	cout << "Fragerizing..." << endl;
	for(DWORD i = 0; i < files.size(); i++){
		HANDLE fileHandle = CreateFileA(files.at(i).c_str(),0,FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE,0,OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0);
		if(fileHandle == INVALID_HANDLE_VALUE){
			cerr << "Error trying to open " << files.at(i) << endl;
			continue; // Maybe this file does not exist anymore, or it never existed
		}
		//Get retrieval pointers.
		RETRIEVAL_POINTERS_BUFFER* retpointbuf = NULL;
		DWORD structSize = sizeof(LARGE_INTEGER) * (2 + 2 * fileClusterCounts[files.at(i)]);
		retpointbuf = (RETRIEVAL_POINTERS_BUFFER*)malloc(structSize);
		STARTING_VCN_INPUT_BUFFER svib;
		svib.StartingVcn.QuadPart = 0;
		DWORD unused;
		if(DeviceIoControl(fileHandle,FSCTL_GET_RETRIEVAL_POINTERS,&svib,sizeof(svib),retpointbuf,structSize,&unused,NULL) == 0){
			if(fileClusterCounts[files.at(i)] > 1)
				cerr << "Error getting retrieval pointers from " << files.at(i) << endl;
			CloseHandle(fileHandle);
			free(retpointbuf);
			continue;
			//die("Failed to get file pointers!");
		}
		ULONGLONG currentVCN = 0;
		for(DWORD j = 0; j < retpointbuf->ExtentCount; j++){
			//Move ... that ... cluster!
			MOVE_FILE_DATA mfdat;
			mfdat.ClusterCount = retpointbuf->Extents[j].NextVcn.QuadPart - currentVCN;
			mfdat.FileHandle = fileHandle;
			mfdat.StartingLcn.QuadPart = currentDiskCluster;
			mfdat.StartingVcn.QuadPart = currentVCN;
			//Try to move
			while(DeviceIoControl(vol,FSCTL_MOVE_FILE,&mfdat,sizeof(mfdat),NULL,0,&unused,NULL) == FALSE){
				cout << "Failed to move " << files.at(i) << " cc: "<< mfdat.ClusterCount 
					<< " vcn: "<<mfdat.StartingVcn.QuadPart<< 
					" currentDiskCluster " << currentDiskCluster << " bitmap->Buffer "
					<< (int)bitmap->Buffer[currentDiskCluster/8] << endl;
				//If that failed because the sector was randomly written into, try to find another contiguous block.
				if(GetLastError() == 5){ //Access denied - should only happen if cluster in use
					printErr();
					cout << "Looking for new cluster set of size " << clustersToReserve << " for remaining files." << endl;
					//Re-get bitmap.
					inputLCN.StartingLcn.QuadPart = 0;
					result = DeviceIoControl(vol, FSCTL_GET_VOLUME_BITMAP, &inputLCN, sizeof(inputLCN), 
						bitmap, bitmapSizeIn, &bitmapSize, NULL);
					if(result != TRUE)
						die("Can't get volume bitmap!");
					
					currentDiskCluster = findFreeClusterSet(bitmap,clustersToReserve);
					mfdat.StartingLcn.QuadPart = currentDiskCluster;
					continue;
				}
				die("Failed to move file!");
			}
			//Take one down, pass it around, n-m clusters of files on the wall!
			currentVCN = retpointbuf->Extents[j].NextVcn.QuadPart;
			currentDiskCluster += mfdat.ClusterCount;
			clustersToReserve -= mfdat.ClusterCount;
		}
		free(retpointbuf);
		CloseHandle(fileHandle);
	}
	cout << "Yay! We're done!" << endl;
	return 0;
}
